
/*
    hdkeyb.c - routines for hdkeyb control
    http://ok1zia.nagano.cz/wiki/Hdkeyb

    Copyright (C) 2010 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

#define PIN_HD_E    0x04
#define PIN_HD_RS   0x01
#define PIN_HD_RW   0x10
#define PIN_HD_D4   0x02
#define PIN_HD_D5   0x80
#define PIN_HD_D6   0x20
#define PIN_HD_D7   0x40
#define PIN_HD_BUSY 0x08

#define PIN_KEYB_A0 0x10
#define PIN_KEYB_A1 0x01
#define PIN_KEYB_A2 0x04



#define PIN(pins, value) if (value) hdkeyb->wr |= (pins); else hdkeyb->wr &= ~(pins);
struct hdkeyb *hdkeyb;

struct hdkeyb *init_hdkeyb(){
    struct hdkeyb *hdkeyb;

    hdkeyb = g_new0(struct hdkeyb, 1);
#ifdef HAVE_LIBFTDI

    hdkeyb->thread = g_thread_create(hdkeyb_main, (gpointer)hdkeyb, TRUE, NULL); 
    if (!hdkeyb->thread) {
        log_addf("Can't create hdkeyb thread\n");
        g_free(hdkeyb);
        return NULL;
    }
#endif    
    return hdkeyb;
}

void free_hdkeyb(struct hdkeyb *hdkeyb){
#ifdef HAVE_LIBFTDI
    if (hdkeyb->thread){
        hdkeyb->thread_break = 1;
        dbg("join hdkeyb...\n");
        g_thread_join(hdkeyb->thread);
        dbg("done\n");
        hdkeyb->thread=0;
    }
#endif
    g_free(hdkeyb);
}


#ifdef HAVE_LIBFTDI
gpointer hdkeyb_main(gpointer xxx){
    struct hdkeyb *hdkeyb;
    int ret, ftdi_vid = 0xa600, ftdi_pid = 0xe114;

    hdkeyb = (struct hdkeyb*)xxx;

    hdkeyb->ftdi = ftdi_new();
    if (!hdkeyb->ftdi){
        log_addf("Can't create hdkeyb ftdi\n");
        return NULL;
    }

    ret = ftdi_usb_open(hdkeyb->ftdi, ftdi_vid, ftdi_pid);
    if (!ret){
        log_addf("Can't open ftdi device %04x:%04x, error=%d %s", ftdi_vid, ftdi_pid, ret, hdkeyb->ftdi->error_str);
        if (ret==-8) log_addf("Maybe run program as root");
        return NULL;
    }
            
    int bitmask = (PIN_HD_RS | PIN_HD_RW | PIN_HD_E | PIN_HD_D4 | PIN_HD_D5 | PIN_HD_D6 | PIN_HD_D7);
    ret=ftdi_set_bitmode(hdkeyb->ftdi, bitmask, BITMODE_SYNCBB);
    dbg("ftdi_set_bitmode(0x%02x)=%d", bitmask, ret);
    if (ret){
        log_addf("Can't enable bitbang, error=%d %s", ret, hdkeyb->ftdi->error_str);
        return NULL;
    }
        
    ret=ftdi_set_baudrate(hdkeyb->ftdi, 1200);
    if (ret){
        log_addf("Can't set baudrate for ftdi, error=%d %s\n", ret, hdkeyb->ftdi->error_str);
        return NULL;
    }

    while(!hdkeyb->thread_break){



    }
    //todo close kde?
    ftdi_set_bitmode(hdkeyb->ftdi, 0x00, BITMODE_RESET);
    return 0;

}


void hdkeyb_read_handler(struct hdkeyb *hdkeyb, char *s){
    struct zstring *zs;
    char *cmd;
    
    zs = zstrdup(s);
    
    cmd = ztokenize(zs, 1);
//     dbg("trig_read_handler   rcvd: '%s'  cmd='%s'\n", s, cmd);
    if (strcmp(cmd, "!")==0){  /* error */
        log_addf("Hdkeyb error: %s", ztokenize(zs, 0));
    }

    if (strcasecmp(cmd, "key")==0){
        log_addf("Hdkeyb key='%c'");
    }

    zfree(zs);
    redraw_later();
}

void hdkeyb_set_qtf(struct hdkeyb *hdkeyb, struct rotar *rotar){

    
}


long int difftimeval_ms(struct timeval *stop, struct timeval *start){
    int sec, usec; 

    usec = stop->tv_usec - start->tv_usec;
    sec = stop->tv_sec - start->tv_sec;
    if (usec < 0){
        usec += 1000000;
        sec --;
    }
    if (sizeof(long) == 4){
        if (sec > 2000000) sec = 2000000;
    }
    if (sizeof(long) == 2){
        if (sec > 63) sec = 63;
    }
    return (long)sec * 1000L + (long)usec / 1000L;
}






void hdkeyb_send(struct hdkeyb *hdkeyb){
//    dbg("hdkeyb_send(%x) slen=%d\n", (unsigned char)hdkeyb->wr, hdkeyb->slen);
    hdkeyb->sbuf[hdkeyb->slen++] = hdkeyb->wr;
    if (hdkeyb->slen==HDKEYB_BUFLEN) hdkeyb_flush(hdkeyb);
}

int hdkeyb_flush(struct hdkeyb *hdkeyb){
    int ret;
    int l;

    if (!hdkeyb->slen) return 0;

    l = hdkeyb->slen;
    hdkeyb->slen = 0;
    hdkeyb->rlen = 0;

//    dbg("hdkeyb_flush() slen=%d\n", l);
    ret = ftdi_write_data(hdkeyb->ftdi, (unsigned char*)hdkeyb->sbuf, l);
    if (ret != l){
        zwrite(tpipe->threadpipe_write, zconcatesc("HD", "!", "ftdi_write_data", hdkeyb->ftdi->error_str));
        return -1;
    }
#if 0
    int i;
    for (i=0; i<l; i++) hdkeyb_debug_pins(hdkeyb->sbuf[i], "f");    
#endif
#if 0
    { int i;
    dbg("w ");
    for (i=0; i<ret; i++) dbg("%02x ", (unsigned char)hdkeyb->sbuf[i]);    
    dbg("\n"); }
#endif

    ret = ftdi_read_data(hdkeyb->ftdi, (unsigned char *)hdkeyb->rbuf, l);
    if (ret != l){
        zwrite(tpipe->threadpipe_write, zconcatesc("HD", "!", "ftdi_read_data", hdkeyb->ftdi->error_str));
        return -1;
    }
#if 0
    for (i=0; i<ret; i++) hdkeyb_debug_pins(hdkeyb->sbuf[i], "f");    
#endif
#if 0
    {int i;
    dbg("r ");
    for (i=0; i<ret; i++) dbg("%02x ", (unsigned char)hdkeyb->sbuf[i]);    
    dbg("\n");}
#endif
    hdkeyb->rlen = ret;

    return 0;
}

int hdkeyb_reset(struct hdkeyb *hdkeyb){
    int ret;
    int d = 1;

    hdkeyb->wr &= ~(PIN_HD_D7|PIN_HD_D6|PIN_HD_D5|PIN_HD_D4|PIN_HD_E|PIN_HD_RW|PIN_HD_RS);
    hdkeyb->wr |= PIN_HD_BUSY;
    
//    dbg("hdkeyb_reset\n");
    hdkeyb_send(hdkeyb);
    hdkeyb_flush(hdkeyb);
    usleep(16000*d); // more than 15 ms

    ret = hdkeyb_cmd_nowait(hdkeyb, 0x3);  // - - - -   0 0 1 DL 
    if (ret) return ret;
    hdkeyb_flush(hdkeyb);
    usleep(5000*d);  // more than 4.1 ms

    ret = hdkeyb_cmd_nowait(hdkeyb, 0x3);
    if (ret) return ret;
    hdkeyb_flush(hdkeyb);
    usleep(1000*d);  // more than 100 us

    ret = hdkeyb_cmd_nowait(hdkeyb, 0x3);
    if (ret) return ret;
    hdkeyb_flush(hdkeyb);
    usleep(6000*d);  

    ret = hdkeyb_cmd_nowait(hdkeyb, 0x2);  // - - - -   0 0 1 DL    datalen
    if (ret) return ret;
    hdkeyb_flush(hdkeyb);
    usleep(10000*d);

    ret = hdkeyb_cmd(hdkeyb, 0x28); // 0 0 1 DL  N F - -     datalen numberlines font
    if (ret) return ret;
    ret = hdkeyb_cmd(hdkeyb, 0x0c); // 0 0 0 0   1 D C B     displayon cursoron blinkcursor
    if (ret) return ret;
    ret = hdkeyb_cmd(hdkeyb, 0x01); // 0 0 0 0   0 0 0 1     clear display, DDRAM addr=0
    if (ret) return ret;
    ret = hdkeyb_cmd(hdkeyb, 0x06); // 0 0 0 0   0 1 I/D S   increment shift
    if (ret) return ret;

    hdkeyb_flush(hdkeyb);
    return 0;
}

int hdkeyb_setdir(struct hdkeyb *hdkeyb){
    int ret;

    int bitmask = (PIN_HD_RS | PIN_HD_RW | PIN_HD_E | PIN_HD_D4 | PIN_HD_D5 | PIN_HD_D6 | PIN_HD_D7);
    ret = ftdi_set_bitmode(hdkeyb->ftdi, bitmask, BITMODE_SYNCBB);
    if (ret){
        zwrite(tpipe->threadpipe_write, zconcatesc("HD", "!", "ftdi_set_bitmode", hdkeyb->ftdi->error_str));
        return ret;
    }
    return ret;
}

int hdkeyb_cmd_nowait(struct hdkeyb *hdkeyb, char c){
//    int ret;  
    
//    dbg("hdkeyb_cmd_nowait(%02x)\n", (unsigned char)c);
    PIN(PIN_HD_RS, 0);
    PIN(PIN_HD_RW, 0);
    PIN(PIN_HD_E, 0);
    hdkeyb_send(hdkeyb);

    PIN(PIN_HD_D4, c & 0x01);
    PIN(PIN_HD_D5, c & 0x02);
    PIN(PIN_HD_D6, c & 0x04);
    PIN(PIN_HD_D7, c & 0x08);
    PIN(PIN_HD_E, 1);
    hdkeyb_send(hdkeyb);

    PIN(PIN_HD_E, 0);
    hdkeyb_send(hdkeyb);

    return 0;
}

int hdkeyb_write(struct hdkeyb *hdkeyb, char c){
//    int ret;

//    dbg("hdkeyb_write(%02x) slen=%d\n", (unsigned char)c, hdkeyb->slen);
    PIN(PIN_HD_RW, 0);
    PIN(PIN_HD_E, 0);
    hdkeyb_send(hdkeyb);

    PIN(PIN_HD_D4, c & 0x10);
    PIN(PIN_HD_D5, c & 0x20);
    PIN(PIN_HD_D6, c & 0x40);
    PIN(PIN_HD_D7, c & 0x80);
    PIN(PIN_HD_E, 1);
    hdkeyb_send(hdkeyb);

    PIN(PIN_HD_E, 0);
    hdkeyb_send(hdkeyb);

    PIN(PIN_HD_D4, c & 0x01);
    PIN(PIN_HD_D5, c & 0x02);
    PIN(PIN_HD_D6, c & 0x04);
    PIN(PIN_HD_D7, c & 0x08);
    PIN(PIN_HD_E, 1);
    hdkeyb_send(hdkeyb);
    
    PIN(PIN_HD_E, 0);
    hdkeyb_send(hdkeyb);

    return 0;
}

int hdkeyb_wait(struct hdkeyb *hdkeyb){
    int busy = 1;
    struct timeval start, stop;
    long int diff;

    //usleep(2000);
    return 0;
    gettimeofday(&start, NULL);
    while(busy){
        PIN(PIN_HD_RS, 0);
        PIN(PIN_HD_RW, 1); // read
        PIN(PIN_HD_E, 0);
        hdkeyb_send(hdkeyb);
        
        PIN(PIN_HD_E, 1);
        hdkeyb_send(hdkeyb);

        PIN(PIN_HD_E, 0);
        hdkeyb_send(hdkeyb);
        
        PIN(PIN_HD_E, 1);
        hdkeyb_send(hdkeyb);

        PIN(PIN_HD_E, 0);
        hdkeyb_send(hdkeyb);
        busy = hdkeyb->rd & PIN_HD_BUSY;

        gettimeofday(&stop, NULL);
        diff = difftimeval_ms(&stop, &start);
        if (diff > 5) {
            error("hdkeyb_wait: no BUSY reply\n");
            break;
        }
    }
    return 0;
}
                                

int hdkeyb_cmd(struct hdkeyb *hdkeyb, char c){
    int ret;
//    dbg("hdkeyb_cmd(%02x)\n", (unsigned char)c);

    PIN(PIN_HD_RS, 0);
    ret = hdkeyb_write(hdkeyb, c);
    if (ret) return ret;

    ret = hdkeyb_wait(hdkeyb);
    if (ret) return ret;
    return 0;
}

int hdkeyb_data(struct hdkeyb *hdkeyb, char c){
    int ret;

//    dbg("hdkeyb_data(%02x)\n", (unsigned char)c);
    PIN(PIN_HD_RS, 1);
    ret = hdkeyb_write(hdkeyb, c);
    if (ret) return ret;

    ret = hdkeyb_wait(hdkeyb);
    if (ret) return ret;

    return 0;
}

void hdkeyb_cgram(struct hdkeyb *hdkeyb, char pos) {
    hdkeyb_cmd(hdkeyb, 0x40|(pos&0x7f));
    return;
}

void hdkeyb_printc(struct hdkeyb *hdkeyb, char *str){
    hdkeyb_setdir(hdkeyb);
    while (*str) {
        hdkeyb_data(hdkeyb, *str);
        str++;
    }
    hdkeyb_flush(hdkeyb);
} 

void hdkeyb_print(struct hdkeyb *hdkeyb, char pos, char *str){
    
    hdkeyb_setdir(hdkeyb);
    hdkeyb_cursor(hdkeyb, pos);
    while(*str){
        hdkeyb_data(hdkeyb, *str);
        str++;
        pos++;
#ifdef MC1601       
        if (pos==0+8) hdkeyb_cursor(hdkeyb, pos=64);
        if (pos==64+8) hdkeyb_cursor(hdkeyb, pos=0);
#endif      
    }
    hdkeyb_flush(hdkeyb);
}

void hdkeyb_clear(struct hdkeyb *hdkeyb){
    unsigned char i;
    
    hdkeyb_cursor(hdkeyb, LINE_1);
    for (i=20;i;i--) {
        hdkeyb_data(hdkeyb, ' ');
    }
    
    hdkeyb_cursor(hdkeyb, LINE_2);
    for (i=20;i;i--) {
        hdkeyb_data(hdkeyb, ' ');
    }
}


/* keyboard routines */
char hdkeyb_keyb_state(struct hdkeyb *hdkeyb){
    int ret, cnt;
    char scan;

    int bitmask = (PIN_HD_RS | PIN_HD_RW | PIN_HD_E );
    ret = ftdi_set_bitmode(hdkeyb->ftdi, bitmask, BITMODE_SYNCBB);
    if (ret){
        zwrite(tpipe->threadpipe_write, zconcatesc("HD", "!", "ftdi_set_bitmode", hdkeyb->ftdi->error_str));
        return 0xff;
    }
    hdkeyb_flush(hdkeyb);
    PIN(PIN_HD_D4,1);
    PIN(PIN_HD_D5,1);
    PIN(PIN_HD_D6,1);
    PIN(PIN_HD_D7,1);
    PIN(PIN_HD_BUSY,1);
    PIN(PIN_KEYB_A0,0);
    PIN(PIN_KEYB_A1,0);
    PIN(PIN_KEYB_A2,0);
    PIN(PIN_HD_E,0);  // already 0
    hdkeyb_send(hdkeyb);

    PIN(PIN_KEYB_A0,1);
    PIN(PIN_KEYB_A1,0);
    hdkeyb_send(hdkeyb);

    PIN(PIN_KEYB_A0,0);
    PIN(PIN_KEYB_A1,1);
    hdkeyb_send(hdkeyb);

    PIN(PIN_KEYB_A0,1);
    PIN(PIN_KEYB_A1,1);
    hdkeyb_send(hdkeyb);
    hdkeyb_send(hdkeyb);
    hdkeyb_flush(hdkeyb);

    cnt = 0;
    scan = 0xff;
    
    if (!(hdkeyb->rbuf[1] & PIN_HD_D4)) { scan = 'A'; cnt++; }
    if (!(hdkeyb->rbuf[1] & PIN_HD_D5)) { scan = '3'; cnt++; }
    if (!(hdkeyb->rbuf[1] & PIN_HD_D6)) { scan = '2'; cnt++; }
    if (!(hdkeyb->rbuf[1] & PIN_HD_D7)) { scan = '1'; cnt++; }

    if (!(hdkeyb->rbuf[2] & PIN_HD_D4)) { scan = 'B'; cnt++; }
    if (!(hdkeyb->rbuf[2] & PIN_HD_D5)) { scan = '6'; cnt++; }
    if (!(hdkeyb->rbuf[2] & PIN_HD_D6)) { scan = '5'; cnt++; }
    if (!(hdkeyb->rbuf[2] & PIN_HD_D7)) { scan = '4'; cnt++; }

    if (!(hdkeyb->rbuf[3] & PIN_HD_D4)) { scan = 'C'; cnt++; }
    if (!(hdkeyb->rbuf[3] & PIN_HD_D5)) { scan = '9'; cnt++; }
    if (!(hdkeyb->rbuf[3] & PIN_HD_D6)) { scan = '8'; cnt++; }
    if (!(hdkeyb->rbuf[3] & PIN_HD_D7)) { scan = '7'; cnt++; }

    if (!(hdkeyb->rbuf[4] & PIN_HD_D4)) { scan = 'D'; cnt++; }
    if (!(hdkeyb->rbuf[4] & PIN_HD_D5)) { scan = '#'; cnt++; }
    if (!(hdkeyb->rbuf[4] & PIN_HD_D6)) { scan = '0'; cnt++; }
    if (!(hdkeyb->rbuf[4] & PIN_HD_D7)) { scan = '*'; cnt++; }
//    dbg("%d %c\n", scan, scan);

    if (cnt != 1) return 0xff;
    return scan;
}

char hdkeyb_key(struct hdkeyb *hdkeyb){
    char state;

    state = hdkeyb_keyb_state(hdkeyb);
    if (state == hdkeyb->oldkeystate) return '\0';
    hdkeyb->oldkeystate = state;
    return state;
}

int hdkeyb_debug_pins(char a, char *t){
    dbg("%s E=%d  RS=%d  RW=%d  D=%d  BUSY=%d\n", 
            t,
            !!(a & PIN_HD_E),
            !!(a & PIN_HD_RS),
            !!(a & PIN_HD_RW),
            (!!(a & PIN_HD_D4))+ 
            (!!(a & PIN_HD_D5))*2+ 
            (!!(a & PIN_HD_D6))*4+ 
            (!!(a & PIN_HD_D7))*8, 
            !!(a & PIN_HD_BUSY));
    return 0;
}

int hdkeyb_test(void){
    struct hdkeyb *hdkeyb;
    int i;

    dbg("\n\n\n");
    hdkeyb = init_hdkeyb();
    if (!hdkeyb) {
        error("init_hdkeyb() failed\n");
        return -1;
    }
    hdkeyb_setdir(hdkeyb);
    hdkeyb_reset(hdkeyb);
    usleep(300000);

/*//    dbg("1 %x\n", hdkeyb->rd);
    hdkeyb->wr = 0x55;
    WRITE;
//    dbg("2 %x\n", hdkeyb->rd);
    hdkeyb->wr = 0xaa;
    WRITE;
//    dbg("3 %x\n", hdkeyb->rd);
    READ;
//    dbg("4 %x\n", hdkeyb->rd);
    READ;
    READ;
//    dbg("5 %x\n", hdkeyb->rd);
    hdkeyb_free(hdkeyb);
    */

/*    while(1){
        int a= hdkeyb_key(hdkeyb);
        if (a==0) continue;
        dbg("%c\n", a);
    } */
//  gst_start();
//    hdkeyb_print2(hdkeyb, 0, "ABC");
    char ss[50];
    strcpy(ss, "");
    for (i=0; i<100000000; i++){
        char s[50];
        sprintf(s, "jedeme %d", i);
        hdkeyb_setdir(hdkeyb);
        hdkeyb_print(hdkeyb, 80, s);
        
        int a= hdkeyb_key(hdkeyb);
        if (a==0) continue;
        if (a==-1) continue;

        dbg("%d\n", a);
        switch(a){
            case 'C':
                if (strlen(ss)==0) break;
                ss[strlen(ss)-1]='\0';
                break;
            default:
                if (strlen(ss)==15) break;
                ss[strlen(ss)+1]='\0';
                ss[strlen(ss)]=a;
                break;
        }
        sprintf(s, "%-16s", ss);
        hdkeyb_print(hdkeyb, 16, s);
        
    }
//  gst_stop();
    return 0;
}
#endif
